"""
# prep.py

Obtain ROCCO conformable input from samples' BAM files.

ROCCO accepts multiple samples' coverage tracks that are assumed to have been generated following comparable
experimental, alignment, QC, etc. protocols. Though ROCCO is generally robust to deviations from this assumption,
we divide the BAM->WIG step and ROCCO itself so that users can, if necessary, apply their own preprocessing
protocol to the alignment files for computing coverage, normalization, etc. as their experimental context
and goals dictate. This general utility for generating ROCCO conformable input (coverage tracks) from
BAM files is offered as part of the ROCCO package as a subcommand `prep` for convenience and has produced
sufficient results across various tested scenarios.

BAM files for each sample in `--bamdir` are used to create fixed-step bigwig signal tracks.
These signal tracks are then split by chromosome into wig files in the subdirectories `tracks_chr[]`.

The resulting `tracks_chr[]` directories can be listed in the `--param_file` parsed by [rocco gwide](https://nolan-h-hamilton.github.io/ROCCO/rocco/gwide.html)

Alternatively, if running [rocco chrom](https://nolan-h-hamilton.github.io/ROCCO/rocco/chrom.html) manually, e.g.,
    ```
    rocco chrom --chrom chr11 --wig_path tracks_chr11
    ```

Parameters:
    -i, --bamdir (str, default='.'):
        Path to the directory containing BAM files.
    -o, --outdir (str, default='.'):
        Output directory.
    -s, --sizes (str, default='hg38'):
        A path to a chromosome sizes file OR an assembly name.
    -L, --interval_length (int, default=50):
        Wiggle track fixed step size
    -c, --cores (int, default=1):
        PEPATAC's `bamSitesToWig.py` cores parameter. Altering this parameter to use >1 core may cause issues on Mac OS.
    --scale_factor_file (str,default=None):
        Path to a tab-separated file with two columns: samples' alignment filenames and the other for their respective scale factors as computed with an external method.
        For instance, an output file from [deepTools](https://deeptools.readthedocs.io/en/develop/content/tools/multiBamSummary.html) `multiBamSummary --scalingFactors` command.

Examples:
    - create input `tracks_chr[]` directories for the hg38 assembly, no sizes file available:
        ```
        rocco prep -s hg38 --bamdir [/path/to/bamfiles]
        ```
    - create input `tracks_chr[]` directories from BAM files using a locally available sizes file:
        ```
        rocco prep -s [/path/to/sizefile] --bamdir [path/to/bamfiles]
        ```
    - create input`tracks_chr[]` directories from hg19 BAM files in the current directory, no sizes file available:
        ```
        rocco prep -s hg19 --bamdir .
        ```

Note:
    External utilities such as [`deepTools bamCoverage`](https://deeptools.readthedocs.io/en/develop/content/tools/bamCoverage.html) can also be used to
    generate track data from BAM files, e.g.,
        ```
        bamCoverage -b sample.bam --binSize 50 -o sample.bam.bw [...]
        ```
        Then, to structure input using UCSC's `bigWigToWig` :
        ```
        bigWigToWig sample.bam.bw tracks_chr[]/chr[]_sample.bam.bw.wig -chrom=chr[]
        ```
    Such utilities offer several additional features that users may find helpful for
    normalization, sample comparison, visualization/plotting, etc. but ROCCO has not
    been tested rigorously on the signals generated from these tools.

"""

import os
import argparse
import subprocess
import pandas as pd
import pysam
from . import rocco_aux
import tempfile
import warnings


def main(args):
    # process command line arguments
    cwd = os.getcwd()
    args['bamdir'] = os.path.abspath(args['bamdir'])
    args['outdir'] = os.path.abspath(args['outdir'])
    sf_dict = None
    if args['scale_factor_file'] is not None:
        aln_names = pd.read_csv(args['scale_factor_file'],sep='\t',header=None)[0]
        scale_factors = pd.read_csv(args['scale_factor_file'],sep='\t',header=None)[1]
        sf_dict = dict(zip(aln_names,scale_factors))

    if not os.path.exists(args['sizes']):
        # if args['sizes'] is not a locally available file,
        # check UCSC for the assembly's sizes file
        try:
            downloaded_size_file = rocco_aux.get_size_file(args['sizes'])
        except Exception as ex:
            print(f'prep: {args["sizes"]} could not be found locally as a file (e.g. `hg38.sizes`) or on UCSC server as an assembly name (e.g. `hg38`).')
            raise
        args['sizes'] = downloaded_size_file

    chroms = list(rocco_aux.parse_size_file(args['sizes']).keys())
    bigwig_files = []
    tf = tempfile.NamedTemporaryFile(mode='w', delete=False)

    for fname in os.listdir(args['bamdir']):
        fname_sf = 1.0
        try:
            if args['scale_factor_file'] is not None:
                fname_sf = float(sf_dict[fname])
        except:
            warnings.warn(
                f'prep: unable to parse scale factor for {fname} in {args["scale_factor_file"]}...setting to 1.0')
            fname_sf = 1.0

        fname = os.path.join(args['bamdir'], fname)
        fname = os.path.abspath(fname)

        # create an index for bam files if not present already
        if rocco_aux.is_alignment(fname):
            aln = pysam.AlignmentFile(fname)
            try:
                aln.check_index()
            except ValueError:
                print(f'prep: no index file available for {fname}, calling pysam.index()')
                pysam.index(fname)

            print('prep: {}'.format(aln.filename.decode()))
            bstw_path = os.path.join(os.path.dirname(__file__), "bamSitesToWig.py") 
            bstw_cmd = ['python3', bstw_path,
                        '-i', aln.filename.decode(),
                        '--scale', str(fname_sf),
                        '-c', args['sizes'],
                        '-w', args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw',
                        '-r', str(args['interval_length']),
                        '-m', 'atac',
                        '-l', str(int(args['interval_length'])//2),
                        '-p', str(args['cores']),
                        '--variable-step',
                        '--limit', " ".join([x for x in chroms if rocco_aux.has_reads(bamfile=aln.filename.decode(),min_reads=1,chrom=x)])]

            if not args['multi']:
                # if `--multi False
                subprocess.run(bstw_cmd,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL,
                               shell=False, check=True)
            else:
                tf.write(' '.join(bstw_cmd))
                tf.write('\n')

            bigwig_file = args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw'
            bigwig_files.append(bigwig_file)
    tf.close()

    if args['multi']:
        rocco_aux.run_par(tf.name)

    # after running bamSitesToWig, split each bigwig file by chromosome into tracks_chr[]
    os.chdir(args['outdir'])
    for chrom in chroms:
        try:
            os.mkdir(f'tracks_{chrom}')
        except FileExistsError:
            pass
        for bigwig_file in [bw_file.split('/')[-1] for bw_file in bigwig_files]:
            subprocess.run(['bigWigToWig', bigwig_file,
                        chrom + '_' + bigwig_file + '.wig',
                        '-chrom={}'.format(chrom)], check=True)

            # if wig file is nonempty, place in tracks_chr[]
            if os.path.getsize(chrom + '_' + bigwig_file + '.wig') > 1024:
                subprocess.run(["mv", chrom + '_' + bigwig_file + '.wig',
                        "tracks_{}".format(chrom)], check=True)
            else:
                subprocess.run(["rm", chrom + '_' + bigwig_file + '.wig'], check=True)

    os.chdir(cwd)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='BAM preprocessing')
    parser.add_argument('-i','--bamdir', default='.', type=str, help='Path to the directory containing BAM files.')
    parser.add_argument('-o', '--outdir', default= '.', help='Output directory')
    parser.add_argument('-s', '--sizes',
                        default='hg38',
                        help="A path to a chromosome sizes file. OR\
                        an assembly name")
    parser.add_argument('-L', '--interval_length',
                        default=50,
                        help="wiggle track fixed step size")
    parser.add_argument('-c', '--cores',
                        type=int,
                        default=1,
                        help="`bamSitesToWig.py`'s cores  parameter.\
                        Altering this parameter to use >1 core\
                        may cause issues on Mac OS")
    parser.add_argument('--scale_factor_file', default=None, help="Path to a tab-separated file with two columns:\
                        the first for samples' alignment filenames and the other for their respective scale factors.\
                        e.g., the output file from deepTools `multiBamSummary --scalingFactors` command.")
    parser.add_argument('--multi',
                         default=True)
    args = vars(parser.parse_args())
    main(args)



"""
# rocco_aux.py

Miscellaneous helper functions

"""
import os
import subprocess
import pandas as pd
import numpy as np
import pysam
import tempfile
import warnings
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor

def sort_combine_bed(outfile: str, dir_: str = '.', exclude_list: list = ['EBV', 'M', 'MT']):
    """
    Sorts and combines chromosome-specific peak files. Creates a new
    bed file `outfile` to store the combined results.

    Parameters:
        outfile (str): output file name
        dir_ (str): directory containing chromosome-specific bed files
            Default is the current directory ('.').
        exclude_list (list): list of chromosomes to exclude. (default: ['EBV', 'M', 'MT'])

    Notes:
        - this function assumes the bed files are named with the\
            same convention and delimiter '_'
    """
    dir_ = os.path.normpath(dir_)

    filenames = [f_ for f_ in os.listdir(dir_)
                 if f_.split('.')[-1] == 'bed' and 'ROCCO_out' in f_
                 and f_ not in exclude_list]

    name_template = []
    try:
        name_template = filenames[0].split('_')
    except (IndexError, ValueError) as none_found:
        print('sort_combine_bed(): no valid ROCCO output files found in `dir_`')
        raise none_found
    chr_index = [i for i in range(len(name_template)) if 'chr' in name_template[i][:3]][0]
    filenames = sorted(filenames,
                       key=lambda x: int(val) if (val := x.split('_')[chr_index][3:]).isnumeric() else ord(val))
    filenames = [os.path.join(dir_, fname) for fname in filenames]
    with open(outfile, mode='w', encoding='utf-8') as outfile_:
        for fname in filenames:
            cat_process = subprocess.Popen(('cat', fname), stdout=outfile_.fileno())
            cat_process.wait()


def download_file(url, output_path):
    "download file at `url` to `output_path` using either curl or wget"
    try:
        subprocess.run(['curl', '-o', output_path, url], check=True)
    except subprocess.CalledProcessError:
        try:
            subprocess.run(['wget', '-O', output_path, url], check=True)
        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to download the file.")


def get_size_file(assembly='hg38', exclude_list=['EBV', 'M', 'MT']) -> str:
    """
    if `assembly` is included in the UC Santa Cruz genome repository
    this function downloads, filters, and sorts the corresponding sizes
    file to `assembly`.sizes. Chromosomes in `exclude_list` are ignored.

    Args:
        assembly (str, optional): name of genome assembly. Defaults to 'hg38'.
        exclude_list (list, optional): list of chromosome names to exclude in sizes file.
            Defaults to ['EBV', 'M', 'MT'].

    Returns:
        str: path to sizes file, `assembly`.sizes
    """
    assembly = assembly.lower()
    url = f"http://hgdownload.soe.ucsc.edu/goldenPath/{assembly}/bigZips/{assembly}.chrom.sizes"
    output_path = f"{assembly}.sizes"
    print(f'\n{url} --> {output_path}\n')
    download_file(url, output_path)
    with open(output_path, 'r') as file_:
        lines = file_.readlines()
    filtered_lines = [line for line in lines if line.strip() and not any(excluded_chr in line for excluded_chr in exclude_list) and '_' not in line]
    sorted_lines = sorted(filtered_lines, key=lambda x: int(x.split('\t')[0][3:]) if x.split('\t')[0][3:].isnumeric() else ord(x.split('\t')[0][3:]))
    sorted_output_path = f"{assembly}.sizes"
    with open(sorted_output_path, 'w') as file_:
        file_.writelines(sorted_lines)
    return sorted_output_path


def parse_size_file(size_file, exclude_list=['EBV', 'M', 'MT']) -> dict:
    """Parse a size file and return a dictionary {chr: size}.

    Args:
        size_file (str): Path to the size file.
        exclude_list (list): list of chromosomes to exclude. (default: ['EBV', 'M', 'MT'])
    Returns:
        dict: a dictionary where chromosome names are keys and their sizes are values.
    """
    try:
        df = pd.read_csv(size_file, sep='\t', header=None)
    except Exception as ex:
        print(f'\nrocco_aux.parse_size_file(): Could not parse {size_file}. Ensure it is a valid sizes file.\n')
        raise
    size_dict = {key: val
                 for key, val in OrderedDict(zip(df.iloc[:, 0], df.iloc[:, 1])).items()
                    if key not in exclude_list}
    return size_dict


def is_alignment(filepath) -> bool:
    """
    Check if a file is an alignment file.

    Args:
        filepath (str): The path to the file.

    Returns:
        bool: True if the file is an alignment file, False otherwise.
    """
    if filepath.split('.')[-1] in ['bam','sam']:
        try:
            pysam.head("-n 1", filepath)
            return True
        except pysam.SamtoolsError:
            return False
    return False

def run_par(cmd_file, threads=-1, verbose=False):
    """
    Runs shell commands in `cmd_file` in parallel

    Args:
        cmd_file (str): file containing newline-separated commands
        threads (int): number of threads to use. if default '-1',
            then the number of threads is set to 1 + os.cpu_count()//2
        verbose (bool): whether to print subprocess output to stdout
    """
    assert os.path.exists(cmd_file), f'supplied `cmd_file` could not be found'
    if threads == -1:
        threads = 1 + os.cpu_count()//2
    with open(cmd_file, 'r') as file:
        commands = file.readlines()

    commands = [cmd.strip() for cmd in commands][::-1]


    def run_command(command, verbose=verbose):
        if not verbose:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            process.communicate()
            return command, process.returncode
        if verbose:
            process = subprocess.Popen(command, shell=True, stderr=subprocess.DEVNULL)
            process.communicate()
            return command, process.returncode

    with ThreadPoolExecutor(max_workers = threads) as executor:
        futures = [executor.submit(run_command, cmd) for cmd in commands]
        results = [future.result() for future in futures]
    if verbose:
        for command, returncode in results[::-1]:
            print(f"cmd: {command}\nretval: {returncode}\n")

def has_reads(bamfile, min_reads=1, chrom='chrY') -> bool:
    """
    Check if a BAM file contains a minimum number of reads for a specific chromosome.

    Args:
        bamfile (str): path to the BAM file
        min_reads (int): minimum read threshold
        chrom (str, optional): chromosome to check for reads

    Returns:
        bool: True if the BAM file contains at least 'min_reads' for 'chrom', False otherwise.
    """
    bamfile = pysam.AlignmentFile(bamfile, "rb")
    has_min_reads = False

    for elem in bamfile.get_index_statistics():
        if elem[0] == chrom and elem[1] >= min_reads:
            has_min_reads = True
            break
    bamfile.close()
    return has_min_reads

def get_locus_size(wig_path, fixedStep=False) -> int:
    """
    Infers step size from replicates` wig files.

    Args:
        wig_path (str) : path to *directory* containing wig files
        fixedStep (bool): whether the wiggle tracks are in fixedStep format. Defaults to `False`.

    Returns:
        int: an integer locus/step size used in the wig files
    """
    if fixedStep:
        return parse_fixedStep_header([os.path.join(wig_path,x) for x in os.listdir(wig_path) if os.path.splitext(x)[-1] == '.wig'][0])[2]
    i = 0
    pos1 = 0
    pos2 = 0
    wig_file = [os.path.join(wig_path,x) for x in os.listdir(wig_path) if os.path.splitext(x)[-1] == '.wig'][0]
    with open(wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            if line.replace('\t', '').replace(' ', '').isnumeric():
                if i == 0:
                    pos1 = int(line.split('\t')[0])
                if i == 1:
                    pos2 = int(line.split('\t')[0])
                if pos2 - pos1 > 0:
                    break
                i += 1
    return pos2 - pos1

def get_start_end(wig_path, fixedStep=False):
    """
    Infers starting and ending nucleotide position common to the
    multiple signal tracks

    To infer a reasonable genomic region over which to run ROCCO,
    this function returns the median starting point and endpoints
    among the multiple samples' signal files in `wig_path`.

    Args:
        wig_path (str): path to *directory* containing wig files
        fixedStep (bool): whether the wiggle tracks are in fixedStep format. Defaults to `False`.

    Returns:
        start (int): median starting position
        end (int): median ending position
    """
    starts = []
    ends = []
    def get_start(track, fixedStep=fixedStep):
        if fixedStep:
            start = parse_fixedStep_header(track)[1]
        with open(track, mode='r', encoding='utf-8') as wig:
            for line in wig:
                line = line.strip()

                if line.replace('\t', '').replace(' ', '').isnumeric():
                    start = int(line.split('\t')[0])
                    return start
        return None

    def get_end(track, fixedStep=fixedStep):
        if fixedStep:
            chrom,start,step = parse_fixedStep_header(track)
            length = num_nonempty_lines(track) - 1
            return start + step*(length-1)

        with subprocess.Popen(["tail", "-n1", track],
                              stdout=subprocess.PIPE) as proc:
            for line in proc.stdout:
                line = line.decode()
                line = line.strip()
                return int(line.split('\t')[0])
        return None

    wig_files = [os.path.join(wig_path,x) for x in os.listdir(wig_path) if os.path.splitext(x)[-1].lower() == '.wig']
    for wig_file in wig_files:
        starts.append(get_start(wig_file))
        ends.append(get_end(wig_file))
    return sorted(starts)[len(starts) // 2], sorted(ends)[len(ends) // 2]


def num_nonempty_lines(filepath):
    with open(filepath, 'r') as file:
        nonempty_lines = sum(1 for line in file if not line.strip().isspace())
    return nonempty_lines


def parse_fixedStep_header(fs_wig):
    with open(fs_wig, 'r') as file:
        i=0
        for line in file:
            if line.startswith("fixedStep"):
                line = line.strip()
                declaration_line = line.split()
                chrom = declaration_line[1].split('=')[1]
                start = int(declaration_line[2].split('=')[1])
                step = int(declaration_line[3].split('=')[1])
                if i > 0:
                    warnings.warn(f'fixedStep header line is not first line of {fs_wig}')
                return chrom,start,step
            i += 1
        return None

def tmp_fixedStep(fs_wig, delim='\t'):
    loci = []
    signal = []
    chrom_ct = 0
    chrom = None
    with open(fs_wig, 'r') as file:
        for line in file:
            if line.startswith("fixedStep"):
                if chrom_ct > 0:
                    warnings.warn(f'only the first chromosome, {chrom}, was extracted from {fs_wig}')
                    break
                line = line.strip()
                declaration_line = line.split()
                chrom = declaration_line[1].split('=')[1]
                start = int(declaration_line[2].split('=')[1])
                step = int(declaration_line[3].split('=')[1])
                chrom_ct += 1
            elif line.strip():
                loci.append(start)
                signal.append(float(line))
                start += step
    tmpname = f'{fs_wig}.vstep.wig'
    with open(tmpname, 'w') as vstep_file:
        for loc,sig in zip(loci,signal):
            vstep_file.write(f'{loc}{delim}{sig}\n')
    return tmpname


def read_wig(chrom_wig_file, start=0, end=10**10, locus_size=50, delim='\t', fixedStep=False):
    r"""
    Processes a chromosome-specific wig file for inclusion in the signal matrix $\mathbf{S}_{chr}$.
    Assumes fixed step-size between nucleotide positions.

    Args:
        wig_file (str): path to wiggle-formatted signal track
        start (int): inferred or manually-specified starting nucleotide
          position
        end (int): inferred or manually-specified ending nucleotide
          position
        locus_size (int): interval/locus size. Each wig file will have
          signal values at every `start + i*locus_size` nucleotide pos.
          for i = 0,1,2,...
        delim (str): delimiter used to separate nucleotide positions from signal values in
            wiggle tracks. Defaults to tabs `\t`
        fixedStep (bool): whether the wiggle track is in fixedStep format. Defaults to `False`.

    Returns:
        loci: list of starting nucleotide positions for each locus
        signal: enrichment signal value at each locus in `loci`
    """

    print("reading: {}".format(chrom_wig_file))
    tmpfile = None
    if fixedStep:
         tmpfile = tmp_fixedStep(chrom_wig_file, delim=delim)
         chrom_wig_file = tmpfile
    loci = []
    signal = []
    with open(chrom_wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            try:
                line = line.split(delim)
                line[0] = int(line[0])
                line[1] = float(line[1])

                if line[0] < start:
                    continue
                if line[0] > end:
                    break

                # case: negative signal value
                # behavior: set signal value to zero
                if line[1] < 0:
                    line[1] = 0

                # case: wig begins after specified `start`
                # behavior: pad beginning with zero signal values
                if start < line[0] and len(loci) == 0:
                    for loc in range(start, line[0], locus_size):
                        loci.append(loc)
                        signal.append(0)
                    loci.append(line[0])
                    signal.append(line[1])
                    continue



                # case: missing data, gap between two loci
                # behavior: fill in gaps with zero signal vals
                if len(loci) > 0 and line[0] - loci[-1] > locus_size:
                    loci.append(loci[-1] + locus_size)
                    signal.append(0)
                    continue

                loci.append(line[0])
                signal.append(line[1])

            except ValueError:
                continue
            except IndexError:
                continue
    # case: wig ends prematurely
    # behavior: pad with zeros
    if loci[-1] < end:
        for loc in range(loci[-1] + locus_size, end + locus_size, locus_size):
            loci.append(loc)
            signal.append(0)
    if fixedStep and tmpfile is not None:
        try:
            os.remove(tmpfile)
        except:
            warnings.warn(f'temporary file {tmpfile} not deleted')
    return loci, signal




import os
import subprocess
import shutil
import pytest

solver = 'CLARABEL'
peakfile_minsize = 100
jaccard_min = .95
rr_iter = '100'
chr19_start = '20000000'
chr19_end = '50000000'
chr19_budget= '0.035'
chr20_start = '20000000'
chr20_end = '50000000'
chr20_budget = '0.035'
filter_min_score = '50'
group_A_bedfile = 'group_A.bed'
group_B_bedfile = 'group_B.bed'

tracks_chrom19_wig = os.path.relpath('data/tracks_chr19/chr19_test.bam.bw.wig.orig')
tracks_chrom19_wig_fs = os.path.relpath('data/tracks_chr19_fixedStep/chr19_test_bam.bw.wig.fs.orig')
tracks_chrom20_wig = os.path.relpath('data/tracks_chr20/chr20_test.bam.bw.wig.orig')

chr19_outfile = f'ROCCO_out_chr19_{chr19_budget}_1.0_0.0_1.0_1.0_1.0.bed'
chr19_ref_file = os.path.relpath('data/ref/ref_chr19.bed')

chr20_outfile = f'ROCCO_out_chr20_{chr20_budget}_1.0_0.0_1.0_1.0_1.0.bed'
chr20_ref_file = os.path.relpath('data/ref/ref_chr20.bed')

gwide_outfile = 'gwide_out.bed'
gwide_outfile_fs = 'gwide_out_fs.bed'
gwide_outdir = os.path.relpath('output/combined')
gwide_outdir_fs = os.path.relpath('output/combined_fs')
gwide_ref_file = os.path.relpath('data/ref/ref_gwide.bed')

metadata_file = 'test_metadata.tsv'
identifiers_file = 'test_identifiers.txt'
identifiers_outdir = 'test_identifiers_outdir'
identifiers_outdir_fs = 'test_identifiers_outdir_fs'

@pytest.mark.regular
def test_sim(var=1, samples=10):
    """
    Add noise to existing wiggle file to generate varying tracks for testing purposes.
    `samples` signal tracks are generated.
    """
    proc1 = subprocess.run(['python', 'sim.py', tracks_chrom19_wig, str(var), str(samples)],
                           capture_output=True, encoding='utf-8')
    proc2 = subprocess.run(['python', 'sim.py', tracks_chrom20_wig, str(var), str(samples)],
                           capture_output=True, encoding='utf-8')
    assert proc1.returncode == 0, f'sim.py failed for {tracks_chrom19_wig}'
    assert proc2.returncode == 0, f'sim.py failed for {tracks_chrom20_wig}'

@pytest.mark.regular
def test_rocco_exists():
    """
    Ensure command line utility 'rocco' is available
    """
    assert shutil.which('rocco') is not None

@pytest.mark.regular
def test_chrom_help():
    """
    Ensure 'rocco chrom -h' returns help message
    """
    # check if return code is 0
    proc = subprocess.run(['rocco', 'chrom', '-h'],capture_output=True, encoding='utf-8')
    assert proc.returncode == 0
    assert 'usage: rocco chrom' in proc.stdout

@pytest.mark.regular
def test_gwide_help():
    """
    Ensure 'rocco gwide -h' returns help message
    """
    # check if return code is 0
    proc = subprocess.run(['rocco', 'gwide', '-h'],capture_output=True, encoding='utf-8')
    assert proc.returncode == 0
    assert 'usage: rocco gwide' in proc.stdout

@pytest.mark.regular
def test_budgets_help():
    """
    Ensure 'rocco budgets -h' returns help message
    """
    # check if return code is 0
    proc = subprocess.run(['rocco', 'budgets', '-h'],capture_output=True, encoding='utf-8')
    assert proc.returncode == 0
    assert 'usage: rocco budgets' in proc.stdout

@pytest.mark.regular
def test_prep_help():
    """
    Ensure 'rocco prep -h' returns help message
    """
    # check if return code is 0
    proc = subprocess.run(['rocco', 'prep', '-h'],capture_output=True, encoding='utf-8')
    assert proc.returncode == 0
    assert 'usage: rocco prep' in proc.stdout

@pytest.mark.regular
def test_run_chrom_chr19():
    """
    Ensure 'rocco chrom' runs successfully for chr19
    """
    # check if command returns exit code '0'
    assert subprocess.run(['rocco', 'chrom','--chrom', 'chr19', '--wig_path',  os.path.dirname(tracks_chrom19_wig),
                           '--budget', chr19_budget, '-N', rr_iter, '--start', chr19_start, '--end', chr19_end,
                           '--solver', solver],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco chrom returncode != 0'
    # check if output file was created properly
    assert chr19_outfile in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(chr19_outfile).st_size > peakfile_minsize

@pytest.mark.regular
def test_comp_ref_chr19():
    """
    Ensure the generated BED files are reasonably consistent with constant references.
    Wont match perfectly due the random simulation, but should exhibit broad similarity
    """
    cmd = ["bedtools", "jaccard", "-a", chr19_outfile , "-b", chr19_ref_file ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_chr19 = float(data.split("\t")[2])
    assert jaccard_chr19 >= jaccard_min

@pytest.mark.regular
def test_run_chrom_chr19_identifiers():
    """
    Ensure 'rocco chrom --identifiers' runs successfully and uses correct subset
    of samples in `test_identifiers.txt`
    """
    # check if command returns exit code '0'
    assert subprocess.run(['rocco', 'chrom','--chrom', 'chr19', '--wig_path',  os.path.dirname(tracks_chrom19_wig),
                           '--budget', chr19_budget, '-N', rr_iter, '--start', '10000000', '--end', '20000000',
                           '--solver', solver, '--identifiers', identifiers_file, '--outdir', identifiers_outdir],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco chrom returncode != 0'
    # check if output file was created in the output directory `identifiers_outdir`
    assert chr19_outfile in os.listdir(identifiers_outdir)
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(os.path.join(identifiers_outdir,chr19_outfile)).st_size > peakfile_minsize

    identifier_tags = []
    id_wig_match_dict = {}
    for line in open(identifiers_file, mode='r', encoding='utf-8'):
        identifier_tags.append(line.strip())
    for line in open(os.path.join(identifiers_outdir, 'chr19_sample_wig_match.log'), mode='r', encoding='utf-8'):
        parsed_line = line.strip()
        parsed_line = parsed_line.split('\t')
        id_wig_match_dict.update({parsed_line[0]: parsed_line[1]})

    for samp_tag in identifier_tags:
        assert samp_tag in id_wig_match_dict.keys(), f'no wig file found for sample {samp_tag}'
    assert len(id_wig_match_dict) == len(identifier_tags), f'a sample was missed or included erroneously:\n{str(id_wig_match_dict)}'
    assert len(list(set(id_wig_match_dict.values()))) == len(list(id_wig_match_dict.values())), f'multiple distinct samples erroneously matching to same wig file:\n{str(id_wig_match_dict)}'

@pytest.mark.regular
def test_run_chrom_chr19_filtered():
    """
    Ensure 'rocco chrom' runs successfully for chr19 with a minimum peak score filter
    """
    # check if command returns exit code '0'
    assert subprocess.run(['rocco', 'chrom','--chrom', 'chr19', '--wig_path',  os.path.dirname(tracks_chrom19_wig),
                           '--budget', chr19_budget, '-N', rr_iter, '--start', chr19_start, '--end', chr19_end,
                           '--solver', solver, '--filter_by_score', filter_min_score],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco chrom returncode != 0'
    # check if output file was created properly
    assert chr19_outfile in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(chr19_outfile).st_size > peakfile_minsize
    with open(chr19_outfile,'r') as filtered_chr19_out:
        for line in filtered_chr19_out:
            line = line.strip()
            if line:
                line = line.split('\t')
                assert float(line[4]) >= float(filter_min_score)

@pytest.mark.regular
def test_run_chrom_chr20():
    """
    Ensure 'rocco chrom' runs successfully for chr20
    """
    # check if command returns exit code '0'
    assert subprocess.run(['rocco', 'chrom', '--chrom', 'chr20', '--wig_path', os.path.dirname(tracks_chrom20_wig),
                            '--budget', chr20_budget, '-N', rr_iter, '--start', chr20_start, '--end', chr20_end,  '--solver', solver],
                          capture_output=True, encoding='utf-8').returncode == 0,'rocco chrom returncode != 0'
    # check if output file was created properly
    assert chr20_outfile in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(chr20_outfile).st_size > peakfile_minsize

@pytest.mark.regular
def test_comp_ref_chr20():
    """
    Ensure the generated BED files are reasonably consistent with constant references.
    Wont match perfectly due the random simulation, but should exhibit broad similarity
    """
    cmd = ["bedtools", "jaccard", "-a", chr20_outfile , "-b", chr20_ref_file ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_chr20 = float(data.split("\t")[2])
    assert jaccard_chr20 >= jaccard_min

@pytest.mark.regular
def test_run_gwide():
    """
    Ensure `rocco gwide -p test_params.csv` runs successfully
    """
    # check if command returns exit code '0'
    assert subprocess.run(["rocco", "gwide", "-p", "test_params.csv","--outdir", gwide_outdir, "--combine", gwide_outfile, "-N", rr_iter,  '--solver', solver],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco gwide returncode != 0'
    # check if output file was created properly
    assert gwide_outfile in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(gwide_outfile).st_size > peakfile_minsize
    chrom_files = [x for x in os.listdir(gwide_outdir) if x.split('.')[-1] == 'bed']
    # ensure outdir contains chromosome-specific BED files
    assert len(chrom_files) == 2

@pytest.mark.regular
def test_comp_ref_gwide():
    """
    Ensure the generated BED files are reasonably consistent with constant references.
    Wont match perfectly due the random simulation, but should exhibit broad similarity
    """
    cmd = ["bedtools", "jaccard", "-a", gwide_outfile, "-b", gwide_ref_file]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_gwide = float(data.split("\t")[2])
    assert jaccard_gwide >= jaccard_min

@pytest.mark.extra
def test_gwide_metadata_and_group_comp():
    """
    Ensure correct splitting of samples by group when calling `--coldata`
    """
    # check if command returns exit code '0'
    assert subprocess.run(["rocco", "gwide", "-p", "test_params.csv",  '--coldata', metadata_file, '--group_col', 'status'],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco gwide --coldata != 0'
    assert os.stat(group_A_bedfile).st_size > peakfile_minsize
    assert os.stat(group_B_bedfile).st_size > peakfile_minsize
    """
    Since the division of samples in the metadata file is arbitrary,
    we can expect consistency between `A` and `B` groups'
    results.
    """
    cmd = ["bedtools", "jaccard", "-a", group_A_bedfile, '-b', group_B_bedfile]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_group = float(data.split("\t")[2])
    assert jaccard_group >= jaccard_min

@pytest.mark.regular
def test_gwide_exclude_chroms():
    """
    Running `rocco gwide -p test_params.csv --exclude_chroms chr19,chr20` should not create
    any new BED files since `test_params.csv` only contains entries for these chromosomes.
    """
    num_beds_orig = len([x for x in os.listdir() if x.split('.')[-1] == 'bed'])
    # check if command returns exit code '0'
    assert subprocess.run(["rocco", "gwide", "-p", "test_params.csv",  '--exclude_chroms', 'chr19,chr20'],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco gwide --exclude_chroms != 0'
    num_beds_curr = len([x for x in os.listdir() if x.split('.')[-1] == 'bed'])
    assert num_beds_orig == num_beds_curr

@pytest.mark.fixedStep
def test_sim_fs_chr19(var=1, samples=10):
    """
    simulate data in fixedStep format for chrom19
    """
    proc1 = subprocess.run(['python', 'sim.py', tracks_chrom19_wig_fs, str(var), str(samples), 'fs'],
                           capture_output=True, encoding='utf-8')
    assert proc1.returncode == 0, f'sim.py failed for {tracks_chrom19_wig_fs}'

@pytest.mark.fixedStep
def test_run_chrom_chr19_fixedStep():
    """
    Ensure 'rocco chrom' runs successfully for chr19 with fixedStep input
    """
    # remove output from previous run if present
    if os.path.exists(chr19_outfile):
        os.remove(chr19_outfile)
    # check if command returns exit code '0'
    proc = subprocess.run(['rocco', 'chrom','--chrom', 'chr19', '--wig_path',  os.path.dirname(tracks_chrom19_wig_fs),
                           '--budget', chr19_budget, '-N', rr_iter, '--start', chr19_start, '--end', chr19_end,
                           '--solver', solver, '--fixedStep', '--verbose'],
                          capture_output=True, encoding='utf-8')
    assert proc.returncode == 0, 'rocco chrom failed'
    # check if output file was created properly
    assert chr19_outfile in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(chr19_outfile).st_size > peakfile_minsize

@pytest.mark.fixedStep
def test_comp_ref_chr19_fs():
    """
    same test as in `comp_ref_chr19` but assuming fixedStep format

    note `chr19_outfile` should exist after running `chrom --fixedStep`
    """
    cmd = ["bedtools", "jaccard", "-a", chr19_outfile , "-b", chr19_ref_file ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_chr19 = float(data.split("\t")[2])
    assert jaccard_chr19 >= jaccard_min


@pytest.mark.fixedStep
def test_run_chrom_chr19_identifiers_fs():
    """
    Ensure 'rocco chrom --identifiers' runs successfully and uses correct subset
    of samples in `test_identifiers.txt`
    """
    # check if command returns exit code '0'
    assert subprocess.run(['rocco', 'chrom','--chrom', 'chr19', '--wig_path',  os.path.dirname(tracks_chrom19_wig_fs),
                           '--budget', chr19_budget, '-N', rr_iter, '--start', '10000000', '--end', '20000000',
                           '--solver', solver, '--identifiers', identifiers_file, '--outdir', identifiers_outdir_fs, '--fixedStep'],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco chrom returncode != 0'
    # check if output file was created in the output directory `identifiers_outdir`
    assert chr19_outfile in os.listdir(identifiers_outdir)
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(os.path.join(identifiers_outdir,chr19_outfile)).st_size > peakfile_minsize

    identifier_tags = []
    id_wig_match_dict = {}
    for line in open(identifiers_file, mode='r', encoding='utf-8'):
        identifier_tags.append(line.strip())
    for line in open(os.path.join(identifiers_outdir, 'chr19_sample_wig_match.log'), mode='r', encoding='utf-8'):
        parsed_line = line.strip()
        parsed_line = parsed_line.split('\t')
        id_wig_match_dict.update({parsed_line[0]: parsed_line[1]})

    for samp_tag in identifier_tags:
        assert samp_tag in id_wig_match_dict.keys(), f'no wig file found for sample {samp_tag}'
    assert len(id_wig_match_dict) == len(identifier_tags), f'a sample was missed or included erroneously:\n{str(id_wig_match_dict)}'
    assert len(list(set(id_wig_match_dict.values()))) == len(list(id_wig_match_dict.values())), f'multiple distinct samples erroneously matching to same wig file:\n{str(id_wig_match_dict)}'

@pytest.mark.fixedStep
def test_run_gwide_fs():
    """
    Ensure `rocco gwide -p test_params.csv` runs successfully
    """
    # check if command returns exit code '0'
    assert subprocess.run(["rocco", "gwide", "-p", "test_params.csv","--outdir", gwide_outdir_fs, "--combine", gwide_outfile_fs, "-N", rr_iter,  '--solver', solver],
                          capture_output=True, encoding='utf-8').returncode == 0, 'rocco gwide returncode != 0'
    # check if output file was created properly
    assert gwide_outfile_fs in os.listdir()
    # ensure output file is nonempty (at least 'x' bytes)
    assert os.stat(gwide_outfile_fs).st_size > peakfile_minsize
    chrom_files = [x for x in os.listdir(gwide_outdir_fs) if x.split('.')[-1] == 'bed']
    # ensure outdir contains chromosome-specific BED files
    assert len(chrom_files) == 2

@pytest.mark.fixedStep
def test_comp_ref_gwide_fs():
    """
    Can use same ref file as in comp_ref_gwide
    """
    cmd = ["bedtools", "jaccard", "-a", gwide_outfile_fs, "-b", gwide_ref_file]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    data = proc.stdout.splitlines()[1]
    jaccard_gwide = float(data.split("\t")[2])
    assert jaccard_gwide >= jaccard_min
